package java.time.chrono;

import java.util.Objects;
import java.time.temporal.UnsupportedTemporalTypeException;
import java.time.temporal.ChronoField;
import java.time.temporal.ChronoUnit;
import java.time.temporal.TemporalUnit;
import java.time.temporal.TemporalAmount;
import java.time.temporal.TemporalField;
import java.io.Serializable;
import java.time.temporal.TemporalAdjuster;
import java.time.temporal.Temporal;

abstract class ChronoLocalDateImpl<D extends ChronoLocalDate> implements ChronoLocalDate, Temporal, TemporalAdjuster, Serializable
{
    private static final long serialVersionUID = 6282433883239719096L;
    
    static <D extends ChronoLocalDate> D ensureValid(Chronology chronology, Temporal temporal) {
        ChronoLocalDate chronoLocalDate = (ChronoLocalDate)temporal;
        if (!chronology.equals(chronoLocalDate.getChronology())) {
            throw new ClassCastException("Chronology mismatch, expected: " + chronology.getId() + ", actual: " + chronoLocalDate.getChronology().getId());
        }
        return chronoLocalDate;
    }
    
    ChronoLocalDateImpl() {
    }
    
    @Override
    public D with(TemporalAdjuster temporalAdjuster) {
        return super.with(temporalAdjuster);
    }
    
    @Override
    public D with(TemporalField temporalField, long n) {
        return super.with(temporalField, n);
    }
    
    @Override
    public D plus(TemporalAmount temporalAmount) {
        return super.plus(temporalAmount);
    }
    
    @Override
    public D plus(long y, TemporalUnit obj) {
        if ((obj instanceof ChronoUnit)) {
            switch ((ChronoUnit)obj) {
                case DAYS: {
                    return this.plusDays(y);
                }
                case WEEKS: {
                    return this.plusDays(Math.multiplyExact(y, 7L));
                }
                case MONTHS: {
                    return this.plusMonths(y);
                }
                case YEARS: {
                    return this.plusYears(y);
                }
                case DECADES: {
                    return this.plusYears(Math.multiplyExact(y, 10L));
                }
                case CENTURIES: {
                    return this.plusYears(Math.multiplyExact(y, 100L));
                }
                case MILLENNIA: {
                    return this.plusYears(Math.multiplyExact(y, 1000L));
                }
                case ERAS: {
                    return this.with(ChronoField.ERA, Math.addExact(this.getLong(ChronoField.ERA), y));
                }
                default: {
                    throw new UnsupportedTemporalTypeException("Unsupported unit: " + obj);
                }
            }
        }
        else {
            return super.plus(y, obj);
        }
    }
    
    @Override
    public D minus(TemporalAmount temporalAmount) {
        return super.minus(temporalAmount);
    }
    
    @Override
    public D minus(long n, TemporalUnit temporalUnit) {
        return super.minus(n, temporalUnit);
    }
    
    abstract D plusYears(long p0);
    
    abstract D plusMonths(long p0);
    
    D plusWeeks(long x) {
        return this.plusDays(Math.multiplyExact(x, 7L));
    }
    
    abstract D plusDays(long p0);
    
    D minusYears(long n) {
        return (n == Long.MIN_VALUE) ? ((ChronoLocalDateImpl)this.plusYears(Long.MAX_VALUE)).plusYears(1L) : this.plusYears(-n);
    }
    
    D minusMonths(long n) {
        return (n == Long.MIN_VALUE) ? ((ChronoLocalDateImpl)this.plusMonths(Long.MAX_VALUE)).plusMonths(1L) : this.plusMonths(-n);
    }
    
    D minusWeeks(long n) {
        return (n == Long.MIN_VALUE) ? ((ChronoLocalDateImpl)this.plusWeeks(Long.MAX_VALUE)).plusWeeks(1L) : this.plusWeeks(-n);
    }
    
    D minusDays(long n) {
        return (n == Long.MIN_VALUE) ? ((ChronoLocalDateImpl)this.plusDays(Long.MAX_VALUE)).plusDays(1L) : this.plusDays(-n);
    }
    
    @Override
    public long until(Temporal obj, TemporalUnit temporalUnit) {
        Objects.requireNonNull(obj, "endExclusive");
        ChronoLocalDate date = this.getChronology().date(obj);
        if ((temporalUnit instanceof ChronoUnit)) {
            switch ((ChronoUnit)temporalUnit) {
                case DAYS: {
                    return this.daysUntil(date);
                }
                case WEEKS: {
                    return this.daysUntil(date) / 7L;
                }
                case MONTHS: {
                    return this.monthsUntil(date);
                }
                case YEARS: {
                    return this.monthsUntil(date) / 12L;
                }
                case DECADES: {
                    return this.monthsUntil(date) / 120L;
                }
                case CENTURIES: {
                    return this.monthsUntil(date) / 1200L;
                }
                case MILLENNIA: {
                    return this.monthsUntil(date) / 12000L;
                }
                case ERAS: {
                    return date.getLong(ChronoField.ERA) - this.getLong(ChronoField.ERA);
                }
                default: {
                    throw new UnsupportedTemporalTypeException("Unsupported unit: " + temporalUnit);
                }
            }
        }
        else {
            Objects.requireNonNull(temporalUnit, "unit");
            return temporalUnit.between(this, date);
        }
    }
    
    private long daysUntil(ChronoLocalDate chronoLocalDate) {
        return chronoLocalDate.toEpochDay() - this.toEpochDay();
    }
    
    private long monthsUntil(ChronoLocalDate chronoLocalDate) {
        if (this.getChronology().range(ChronoField.MONTH_OF_YEAR).getMaximum() != 12L) {
            throw new IllegalStateException("ChronoLocalDateImpl only supports Chronologies with 12 months per year");
        }
        return (chronoLocalDate.getLong(ChronoField.PROLEPTIC_MONTH) * 32L + (long)chronoLocalDate.get(ChronoField.DAY_OF_MONTH) - (this.getLong(ChronoField.PROLEPTIC_MONTH) * 32L + (long)this.get(ChronoField.DAY_OF_MONTH))) / 32L;
    }
    
    @Override
    public boolean equals(Object o) {
        return this == o || (o instanceof ChronoLocalDate && this.compareTo((ChronoLocalDate)o) == 0);
    }
    
    @Override
    public int hashCode() {
        long epochDay = this.toEpochDay();
        return this.getChronology().hashCode() ^ (int)(epochDay ^ epochDay >>> 32);
    }
    
    @Override
    public String toString() {
        long long1 = this.getLong(ChronoField.YEAR_OF_ERA);
        long long2 = this.getLong(ChronoField.MONTH_OF_YEAR);
        long long3 = this.getLong(ChronoField.DAY_OF_MONTH);
        StringBuilder sb = new StringBuilder(30);
        sb.append(this.getChronology().toString()).append(" ").append(this.getEra()).append(" ").append(long1).append((long2 < 10L) ? "-0" : "-").append(long2).append((long3 < 10L) ? "-0" : "-").append(long3);
        return sb.toString();
    }
}
